@@ -121,10 +121,11 @@ where
 import Data.List (intersperse, partition)
 import Control.Monad (when, unless)
 import Data.Version (showVersion)
+import qualified Data.Version as DV
 import System.Console.GetOpt
                   (ArgOrder(..), OptDescr(..), ArgDescr(..), usageInfo, getOpt)
 import qualified System.FilePath as FilePath
-                  (takeExtension, dropExtension, takeBaseName)
+                  (takeDirectory, takeExtension, dropExtension)
 import System.FilePath ((<.>), (</>), splitSearchPath)
 import System.IO (stderr, openFile, IOMode(..))
 import System.IO.Error (ioeGetErrorString, ioeGetFileName)
@@ -135,9 +136,10 @@ import C2HS.State  (CST, runC2HS, fatal, fatalsHandledBy,
                    SwitchBoard(..), Traces(..), setTraces,
                    traceSet, setSwitch, getSwitch, putTraceStr)
 import qualified System.CIO as CIO
-import C2HS.C     (hsuffix, isuffix, loadAttrC)
-import C2HS.CHS   (loadCHS, dumpCHS, hssuffix, chssuffix, dumpCHI)
+import C2HS.C     (csuffix, hsuffix, isuffix, loadAttrC)
+import C2HS.CHS   (loadCHS, dumpCHS, hssuffix, chssuffix, dumpCHI, hasNonGNU)
 import C2HS.Gen.Header  (genHeader)
+import C2HS.Gen.Wrapper  (genWrappers)
 import C2HS.Gen.Bind      (expandHooks)
 import C2HS.Version    (versnum, version, copyright, disclaimer)
 import C2HS.Config (cppopts, libfname, PlatformSpec(..),
@@ -178,8 +180,6 @@ errTrailer = "Try the option `--help' on its own for more information.\n"
 --
 data Flag = CPPOpts  String     -- ^ additional options for C preprocessor
           | CPP      String     -- ^ program name of C preprocessor
-          | NoGNU               -- ^ suppress GNU preprocessor symbols
-          | NoBlocks            -- ^ suppress MacOS __BLOCKS__ preproc. symbol
           | Dump     DumpType   -- ^ dump internal information
           | Help                -- ^ print brief usage information
           | Keep                -- ^ keep the .i file
@@ -211,14 +211,6 @@ options  = [
          ["cpp"]
          (ReqArg CPP "CPP")
          "use executable CPP to invoke C preprocessor",
-  Option ['n']
-         ["no-gnu"]
-         (NoArg NoGNU)
-         "suppress GNU preprocessor symbols",
-  Option ['b']
-         ["no-blocks"]
-         (NoArg NoBlocks)
-         "suppress MacOS __BLOCKS__ preprocessor symbol",
   Option ['d']
          ["dump"]
          (ReqArg dumpArg "TYPE")
@@ -279,7 +271,7 @@ compile  =
   do
     setup
     cmdLine <- CIO.getArgs
-    case getOpt RequireOrder options cmdLine of
+    case getOpt Permute options cmdLine of
       (opts, []  , [])
         | noCompOpts opts -> doExecute opts Nothing
       (opts, args, [])    -> case parseArgs args of
@@ -330,7 +322,8 @@ compile  =
             fnMsg = case ioeGetFileName err of
                        Nothing -> ""
                        Just s  -> " (file: `" ++ s ++ "')"
-        CIO.hPutStrLn stderr (msg ++ fnMsg)
+        name <- CIO.getProgName
+        CIO.hPutStrLn stderr $ concat [name, ": ", msg, fnMsg]
         CIO.exitWith $ CIO.ExitFailure 1
 
 -- | set up base configuration
@@ -370,19 +363,11 @@ execute opts args | Help `elem` opts = help
         let bndFileWithoutSuffix  = FilePath.dropExtension bndFile
         computeOutputName bndFileWithoutSuffix
         process headerFiles bndFileWithoutSuffix
-          `fatalsHandledBy` die
       Nothing ->
         computeOutputName "."   -- we need the output name for library copying
     copyLibrary
-      `fatalsHandledBy` die
   where
     atMostOne = (foldl (\_ x -> [x]) [])
-    --
-    die ioerr =
-      do
-        name <- CIO.getProgName
-        CIO.putStr $ name ++ ": " ++ ioeGetErrorString ioerr ++ "\n"
-        CIO.exitWith $ CIO.ExitFailure 1
 
 -- | emit help message
 --
@@ -403,8 +388,6 @@ help =
 processOpt :: Flag -> CST s ()
 processOpt (CPPOpts  cppopt ) = addCPPOpts  [cppopt]
 processOpt (CPP      cpp    ) = setCPP      cpp
-processOpt (NoGNU           ) = setNoGNU
-processOpt (NoBlocks        ) = setNoBlocks
 processOpt (Dump     dt     ) = setDump     dt
 processOpt (Keep            ) = setKeep
 processOpt (Library         ) = setLibrary
@@ -465,16 +448,6 @@ addCPPOpts opts  = setSwitch $ \sb -> sb {cppOptsSB = cppOptsSB sb ++ opts}
 setCPP       :: FilePath -> CST s ()
 setCPP fname  = setSwitch $ \sb -> sb {cppSB = fname}
 
--- | set flag to suppress GNU preprocessor symbols
---
-setNoGNU :: CST s ()
-setNoGNU  = setSwitch $ \sb -> sb {noGnuSB = True}
-
--- | set flag to suppress MacOS __BLOCKS__ preprocessor symbols
---
-setNoBlocks :: CST s ()
-setNoBlocks = setSwitch $ \sb -> sb {noBlocksSB = True}
-
 -- set the given dump option
 --
 setDump         :: DumpType -> CST s ()
@@ -551,7 +524,21 @@ process headerFiles bndFile  =
     --
     (chsMod , warnmsgs) <- loadCHS bndFile
     CIO.putStr warnmsgs
-    traceCHSDump chsMod
+    --
+    -- get output directory and create it if it's missing
+    --
+    outFName <- getSwitch outputSB
+    outDir   <- getSwitch outDirSB
+    let outFPath = outDir </> outFName
+    CIO.createDirectoryIfMissing True $ FilePath.takeDirectory outFPath
+    --
+    -- dump the binding file when demanded
+    --
+    flag <- traceSet dumpCHSSW
+    when flag $ do
+      let chsName = outFPath <.> "dump"
+      CIO.putStrLn $ "...dumping CHS to `" ++ chsName ++ "'..."
+      dumpCHS chsName chsMod False
     --
     -- extract CPP and inline-C embedded in the .chs file (all CPP and
     -- inline-C fragments are removed from the .chs tree and conditionals are
@@ -563,37 +550,35 @@ process headerFiles bndFile  =
     -- create new header file, make it #include `headerFile', and emit
     -- CPP and inline-C of .chs file into the new header
     --
-    outFName <- getSwitch outputSB
-    outDir   <- getSwitch outDirSB
     let newHeader     = outFName <.> chssuffix <.> hsuffix
         newHeaderFile = outDir </> newHeader
-        preprocFile   = FilePath.takeBaseName outFName <.> isuffix
+        preprocFile   = outFPath <.> isuffix
     CIO.writeFile newHeaderFile $ concat $
+      [ "#define C2HS_MIN_VERSION(mj,mn,rv) " ++
+        "((mj)<=C2HS_VERSION_MAJOR && " ++
+        "(mn)<=C2HS_VERSION_MINOR && " ++
+        "(rv)<=C2HS_VERSION_REV)\n" ] ++
       [ "#include \"" ++ headerFile ++ "\"\n"
       | headerFile <- headerFiles ]
       ++ header'
-    --
-    -- Check if we can get away without having to keep a separate .chs.h file
-    --
-    case headerFiles of
-      [headerFile] | null header
-        -> setHeader headerFile    -- the generated .hs file will directly
-                                   -- refer to this header rather than going
-                                   -- through a one-line .chs.h file.
-      _ -> setHeader newHeader
+    setHeader newHeader
     --
     -- run C preprocessor over the header
     --
     cpp      <- getSwitch cppSB
     cppOpts  <- getSwitch cppOptsSB
-    noGnu    <- getSwitch noGnuSB
-    noBlocks <- getSwitch noBlocksSB
-    let noGnuOpts =
-          if noGnu
-          then ["-U__GNUC__", "-U__GNUC_MINOR__", "-U__GNUC_PATCHLEVEL__"]
+    let nonGNUOpts =
+          if hasNonGNU chsMod
+          then [ "-U__GNUC__"
+               , "-U__GNUC_MINOR__"
+               , "-U__GNUC_PATCHLEVEL__" ]
           else []
-        noBlocksOpts = if noBlocks then ["-U__BLOCKS__"] else []
-        args = cppOpts ++ noGnuOpts ++ noBlocksOpts ++ [newHeaderFile]
+        [versMajor, versMinor, versRev] = map show $ DV.versionBranch versnum
+        versionOpt = [ "-DC2HS_VERSION_MAJOR=" ++ versMajor
+                     , "-DC2HS_VERSION_MINOR=" ++ versMinor
+                     , "-DC2HS_VERSION_REV=" ++ versRev ]
+        args = cppOpts ++ nonGNUOpts ++ ["-U__BLOCKS__"] ++
+               versionOpt ++ [newHeaderFile]
     tracePreproc (unwords (cpp:args))
     exitCode <- CIO.liftIO $ do
       preprocHnd <- openFile preprocFile WriteMode
@@ -622,22 +607,21 @@ process headerFiles bndFile  =
     --
     -- expand binding hooks into plain Haskell
     --
-    (hsMod, chi, hooksMsgs) <- expandHooks cheader strippedCHSMod
+    (hsMod, chi, wrappers, hooksMsgs) <- expandHooks cheader strippedCHSMod
     CIO.putStr hooksMsgs
     --
     -- output the result
     --
-    dumpCHS (outDir </> outFName) hsMod True
-    dumpCHI (outDir </> outFName) chi           -- different suffix will be appended
+    dumpCHS outFPath hsMod True
+    dumpCHI outFPath chi           -- different suffix will be appended
+    --
+    -- create new wrapper file if necessary
+    --
+    when (not $ null wrappers) $ do
+      wrapper' <- genWrappers wrappers
+      let newWrapperFile = outDir </> outFName <.> chssuffix <.> csuffix
+      CIO.writeFile newWrapperFile $ concat $
+        [ "#include \"" ++ newHeader ++ "\"\n" ] ++ wrapper'
   where
     tracePreproc cmd = putTraceStr tracePhasesSW $
                          "Invoking cpp as `" ++ cmd ++ "'...\n"
-    traceCHSDump mod' = do
-                         flag <- traceSet dumpCHSSW
-                         when flag $
-                           (do
-                              CIO.putStr ("...dumping CHS to `" ++ chsName
-                                         ++ "'...\n")
-                              dumpCHS chsName mod' False)
-
-    chsName = FilePath.takeBaseName bndFile <.> "dump"
